
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "board_view.h"
#include "logical_board.h"


typedef struct _menu_item
{
    char *label;
    WidgetClass *class;
    void (*callback)();
    XtPointer callback_data;
} MenuItem;



/* motif globals */

XtAppContext app_context;
Widget top_level;

Atom WM_DELETE_WINDOW;
Widget transientShell1;
Widget RowColumn1;
Widget Text1;
Widget Frame5;
Widget RowColumn10;
Widget Frame6;
Widget RowColumn8;
Widget ToggleButton4;
Widget ToggleButton5;
Widget Frame7;
Widget RowColumn9;
Widget PushButton4;
Widget PushButton6;

Widget menu_bar;


/* globals */

Logical_Board *lb;
int _argc;
char **_argv;


/* funtions */

void ready_top_level (void);
void ready_new_view_dialog (void);
Widget BuildPulldownMenu (Widget parent, char *menu_title, Boolean tear_off, MenuItem *items);


/****************/


int gXErrorFlag = 0;

int HandleXError(dpy, event)
     Display *dpy;
     XErrorEvent *event;
{
    int a = 0, b = 1;
    b /= a; /* floating exception! */

#if 0
  gXErrorFlag = 1;

  return 0;
#endif

    return 0;
}

void InstallXErrorHandler(Display *display)
{
  int HandleXError();

  XSetErrorHandler(HandleXError);
  XFlush(display);
}

void DeInstallXErrorHandler(Display *display)
{
  XSetErrorHandler(NULL);
  XFlush(display);
}


/****************/


int main(int argc, char *argv[])
{
    char *fallbacks[] = {"*background:#bfbfbfbfbfbf",
			 "*XmPushButton*background:#bfbfbfbfbfbf",
			 NULL};

    srand(time(0));

    _argc = argc;
    _argv = argv;

    top_level = XtVaAppInitialize (&app_context,
				   "PhaserChess",
				   NULL, 0,
				   &argc, argv,
				   fallbacks,
				   XtNtitle, "Phaser Chess",
				   NULL);



    /* to make editres happy... */
    XtAddEventHandler(top_level, (EventMask)0, True, (XtEventHandler)_XEditResCheckMessages, NULL);

    lb = [[Logical_Board alloc] init_logical_board];
    [lb load_file : STARTUP_PCS];

    ready_top_level ();

    XtAppMainLoop(app_context);

    return 0;
}



void quit (Widget w, XtPointer client_data, XtPointer call_data)
{
    exit (0);
}


void add_view (Widget w, XtPointer client_data, XtPointer call_data)
{
    char *display_name;
    Board_View *bv;
    Player_Color pc;
    Boolean red_val;
    Boolean green_val;

    display_name = XmTextGetString (Text1);

    if (strlen (display_name) == 0)
	return;


    XtVaGetValues (ToggleButton4, XmNset, &red_val, NULL);
    XtVaGetValues (ToggleButton5, XmNset, &green_val, NULL);

    printf ("got %d, %d for red and green values.\n", red_val, green_val);


    switch (2 * (int) red_val + (int) green_val)
    {
    case 0:
	pc = C_neutral;
	break;
    case 1:
	pc = C_green;
	break;
    case 2:
	pc = C_red;
	break;
    case 3:
	pc = C_both;
	break;
    default:
	fprintf (stderr, "This can't happen.\n");
	break;
    }


    bv = [[Board_View alloc] init_board
			     : _argc : _argv
			     : app_context
			     : display_name
			     : pc];

    XtFree (display_name);

    XtPopdown (transientShell1);

    [lb add_board_view : bv];
}



void popup_add_view_dialog (Widget w, XtPointer client_data, XtPointer call_data)
{
     XtPopup (transientShell1, XtGrabNone);
}




void ready_top_level (void)
{
    MenuItem file_items[] = {{"Quit", &xmPushButtonWidgetClass, quit, NULL},
			     {NULL}};
    MenuItem view_items[] = {{"New View", &xmPushButtonWidgetClass, popup_add_view_dialog, (XtPointer) lb},
			     {"Delete View", &xmPushButtonWidgetClass, NULL, NULL},
			     {NULL}};


    menu_bar = XmCreateMenuBar (top_level, "menubar", NULL, 0);
    BuildPulldownMenu (menu_bar, "File", True, file_items);
    BuildPulldownMenu (menu_bar, "View", True, view_items);


    XtManageChild (menu_bar);

    ready_new_view_dialog ();

    XtRealizeWidget (top_level);

    InstallXErrorHandler(XtDisplay (top_level));
}


void wm_delete_nvd (Widget w, XtPointer client_data, XtPointer callData)
{
    XtPopdown (transientShell1);
}



void ready_new_view_dialog (void)
{
    WM_DELETE_WINDOW = XmInternAtom(XtDisplay(top_level), "WM_DELETE_WINDOW", False);

    transientShell1 = XtVaCreatePopupShell("transientShell1",
					   transientShellWidgetClass,
					   top_level,
					   XmNtitle, "New Board View",
					   XmNdeleteResponse, XmDO_NOTHING,
					   NULL );

    XmAddWMProtocolCallback(transientShell1, WM_DELETE_WINDOW, wm_delete_nvd, NULL);
    XtAddEventHandler(transientShell1, (EventMask)0, True, (XtEventHandler)_XEditResCheckMessages, NULL);

    
    RowColumn1 = XtVaCreateWidget("RowColumn1",
				  xmRowColumnWidgetClass,
				  transientShell1,
				  NULL );

    Text1 = XtVaCreateWidget("Text1",
			     xmTextWidgetClass,
			     RowColumn1,
			     NULL );



    { /***/
	char *display;
	char *value;

	display = getenv ("DISPLAY");
	if (display == NULL)
	    value = "";
	else
	    value = strdup (display);

	XtVaSetValues (Text1,
		       XmNvalue, value,
		       NULL);

	free (value);
    } /***/


    XtManageChild(Text1);

    Frame5 = XtVaCreateWidget("Frame5",
			      xmFrameWidgetClass,
			      RowColumn1,
			      NULL );


    RowColumn10 = XtVaCreateWidget("RowColumn10",
				   xmRowColumnWidgetClass,
				   Frame5,
				   XmNorientation, XmHORIZONTAL,
				   NULL );


    Frame6 = XtVaCreateWidget("Frame6",
			      xmFrameWidgetClass,
			      RowColumn10,
			      NULL );

    RowColumn8 = XtVaCreateWidget("RowColumn8",
				  xmRowColumnWidgetClass,
				  Frame6,
				  NULL );

    ToggleButton4 = XtVaCreateWidget("Red Controlable",
				     xmToggleButtonWidgetClass,
				     RowColumn8,
				     XmNtitle, "Red Controlable",
				     NULL );
    XtManageChild(ToggleButton4);

    ToggleButton5 = XtVaCreateWidget("Green Controlable",
				     xmToggleButtonWidgetClass,
				     RowColumn8,
				     NULL );
    XtManageChild(ToggleButton5);

    XtManageChild(RowColumn8);
    XtManageChild(Frame6);
    Frame7 = XtVaCreateWidget("Frame7",
			      xmFrameWidgetClass,
			      RowColumn10,
			      NULL );

    RowColumn9 = XtVaCreateWidget("RowColumn9",
				  xmRowColumnWidgetClass,
				  Frame7,
				  NULL );
    XtVaSetValues(RowColumn9,
		  XtVaTypedArg, XmNnumColumns, XmRString,
		  "2", 2,
		  XtVaTypedArg, XmNpacking, XmRString,
		  "XmPACK_COLUMN", 14,
		  NULL );

    PushButton4 = XtVaCreateWidget("Open X Display",
				   xmPushButtonWidgetClass,
				   RowColumn9,
				   NULL );
    XtManageChild(PushButton4);

    PushButton6 = XtVaCreateWidget("Cancel",
				   xmPushButtonWidgetClass,
				   RowColumn9,
				   NULL );
    XtManageChild(PushButton6);

    XtManageChild(RowColumn9);
    XtManageChild(Frame7);
    XtManageChild(RowColumn10);
    XtManageChild(Frame5);
    XtManageChild(RowColumn1);


    XtAddCallback(PushButton6, XmNactivateCallback, wm_delete_nvd, NULL);
    XtAddCallback(PushButton4, XmNactivateCallback, add_view, NULL);
}




Widget BuildPulldownMenu (Widget parent,
			  char *menu_title,
			  Boolean tear_off,
			  MenuItem *items)
{
    Widget PullDown, cascade, widget;
    int i;
    XmString str;

    PullDown = XmCreatePulldownMenu (parent, "_pulldown", NULL, 0);

    if (tear_off) XtVaSetValues (PullDown, XmNtearOffModel, XmTEAR_OFF_ENABLED, NULL);

    str = XmStringCreateLocalized (menu_title);
    cascade = XtVaCreateManagedWidget (menu_title,
				       xmCascadeButtonWidgetClass, parent,
				       XmNsubMenuId, PullDown,
				       XmNlabelString, str,
				       NULL);
    XmStringFree (str);

    for (i=0; items[i].label != NULL; i++)
    {
	widget = XtVaCreateManagedWidget (items[i].label,
					  *items[i].class, PullDown,
					  NULL);

	if (items[i].callback)
	    XtAddCallback (widget, XmNactivateCallback, items[i].callback, items[i].callback_data);
    }

    return cascade;
}
