%{
	#include "pixmap_reader.h"


	int quote_number;
	int width,height,num_cols,chars_per_color,hot_x,hot_y;
	int cur_col;
	char **colors;
	char *name;
	char vis_type;
	char *default_color;
	Display *display;
	Drawable drawable;
	Pixmap pixmap;
	XColor *xcolors;
	int pix_x, pix_y;
%}


ws			[\t\n ]*
quote			\".*\"
comment			\/\*[.\n]*\*\/
word			[a-zA-Z_]+
aftername		{ws}\[{ws}\]{ws}={ws}\{{ws}

%%

%{
	char *p, *q;
	int lop;
	int parse_ret;
%}

{quote}			{
				if (quote_number == 0)
				{
					sscanf(yytext,"\"%d %d %d %d %d %d",
						&width, &height, &num_cols,
						&chars_per_color, &hot_x, &hot_y);

#if 0
					printf("numbers: %d %d %d %d %d %d\n",
						width, height, num_cols,
						chars_per_color, hot_x, hot_y);
#endif

					colors = (char **)malloc(sizeof(char *)*num_cols);
					xcolors = (XColor *)malloc(sizeof(XColor)*num_cols);

					for (lop=0; lop<num_cols; lop++)
						colors[lop] = (char *)malloc(chars_per_color+1);

					cur_col = 0;

					pixmap = XCreatePixmap(display,
								drawable,
								width,
								height,
								DefaultDepth(display,
									     DefaultScreen(display)));
				}
				else if (quote_number < num_cols+1)
				{
					if (cur_col >= num_cols)
					{
						fprintf(stderr,"more colors that reported in pixmap '%s'\n",
							name);
					}
					else
					{
						for (lop=0; lop<chars_per_color; lop++)
							colors[cur_col][lop] = yytext[lop+1];

						colors[cur_col][lop] = '\0';


						p = &yytext[lop+1];
						p[ strlen(p)-1 ] = '\0'; /* kill ending quote */


						parse_ret = 0;
						for (;;)
						{
							while ((*p) == ' ' || (*p) == '\t')
								p++;

							if ((*p) == '\0' || (*p) == '\n')
								break;

							if ((*p) == vis_type)
							{
								/* read in the color */
								p++;
								while ((*p) == ' ' || (*p) == '\t')
									p++;

								if ((*p) == '\0' || (*p) == '\n')
								{
									fprintf(stderr,"error, missing color name\n");
								}

								q=p;
								while ((*q) != ' ' &&
									(*q) != '\t' &&
									(*q) != '\0' &&
									(*q) != '\n')
									q++;

								(*q) = '\0';
								parse_ret = XParseColor(display,
											DefaultColormap(display,
												DefaultScreen(display)),
											p,
											&xcolors[cur_col]);
								break;
							}
							else
							{
								/* skip the color */
								p++;
								while ((*p) == ' ' || (*p) == '\t')
									p++;
								while ((*p) != ' ' &&
									(*p) != '\t' &&
									(*p) != '\0' &&
									(*p) != '\n')
									p++;
							}
						}

						if (parse_ret == 0)
						{
							fprintf(stderr,"color lookup failed\n");
							parse_ret = XParseColor(display,
										DefaultColormap(display,
												DefaultScreen(display)),
										default_color,
										&xcolors[cur_col]);
						}


						XAllocColor (display,
								DefaultColormap(display, DefaultScreen(display)),
								&xcolors[cur_col]);

						cur_col++;
					}
				}
				else
				{
					p = yytext + 1; /* skip quote */
					q = yytext + strlen(yytext) - 1; /* skip quote */

					while (p<q)
					{
						for (lop=0; lop<num_cols; lop++)
						{
							if (strncmp(colors[lop],p,chars_per_color) == 0)
							{
								break;
							}
						}
						if (lop == num_cols)
						{
							fprintf(stderr, "unknown color: --");
							for (lop=0; lop<chars_per_color; lop++)
								fprintf(stderr,"%c", *(p+lop));
							fprintf(stderr,"--\n");
							lop = 0;
						}

						XSetForeground(display,
								DefaultGC(display, DefaultScreen(display)),
								xcolors[lop].pixel);
						XDrawPoint(display, pixmap,
								DefaultGC(display, DefaultScreen(display)),
								pix_x, pix_y);
						pix_x++;
						p+=chars_per_color;
					}
					pix_x = 0;
					pix_y++;
				}
				quote_number++;
			}

{comment}		;

{word}			;

{word}/{aftername}	{
				/* printf ("NAME:--%s--\n",yytext); */
				name = (char *)malloc(sizeof(char)*(strlen(yytext)+1));
				strcpy(name,yytext);
			}

{aftername}		;

\*			;
\/			;
\}			;
\;			;
\,			;
{ws}			;

%%



Pixmap load_pixmap(Display *set_display, Drawable set_drawable,
			char *filename, char set_vis_type, char *set_default_color)
{
	int lop;

	quote_number = 0;
	pix_x = 0;
	pix_y = 0;


	yyin = fopen(filename,"r");

	if (yyin == NULL)
	{
		perror("fopen");
		fprintf(stderr,"cannot open file '%s' for reading.\n",filename);
		return (Pixmap)0;
	}
	else
	{
		vis_type = set_vis_type;
		default_color = set_default_color;
		display = set_display;
		drawable = set_drawable;

		yylex();
	}


	/* free everything */

	for (lop=0; lop<num_cols; lop++)
		free(colors[lop]);

	free(colors);
	free(xcolors);
	free(name);


	fclose(yyin);

	return pixmap;
}
